@@ -26,46 +26,73 @@
 // custom librairies
 #include "gps.h"
 #include "services.h"
-#include "gateway.h"
+#include "mac-arp.h"
 #include "timeout.h"
 #include "fwrd.h"
+#include "route.h"
+#include "src-ip.h"
 
 
 int nbopt=0, nbports, nb_src_ip=0, rand_seed, sockets[2];
 int ping_port=80, fragment=NO_FRAG, spoofing=MAC_SPOOF, scan_type=SYN_SCAN;
 u_int32_t dst_ip=0;
-services_t *st;
-u_int16_t *tab_response;
-fwrd_elt *tab_fwrd;
-arp_elt *src_ip;
-port_elt *tab_port;
+services_t *st=NULL;
+u_int16_t *tab_response=NULL;
+fwrd_elt *tab_fwrd=NULL;
+arp_elt *src_ip=NULL;
+port_elt *tab_port=NULL;
+devlist *dl=NULL;
+netdev dev;
+struct pcap *descr=NULL;
 u_char dst_mac[ETH_ALEN];
-char errbuf[GPS_ERRBUF_SIZE], *eth=NULL;
+char errbuf[GPS_ERRBUF_SIZE];
 extern char *optarg;
 
 
+// signal handling
+void immune2sig()
+{
+	int i;
+	for(i=0; i<SIGCHLD; i++) signal(i, SIG_IGN);
+}
+void common_sig(int sig)
+{
+	fprintf(stderr, "): caught signal %d - exiting\n", sig);
+	exit(sig);
+}
 void main_sig(int sig)
 {
-	printf("main: dead %d\n", sig);
-	exit(0);
+	immune2sig();
+	if(descr) pcap_close(descr);
+	fprintf(stderr, "Sniffer (parent");
+	switch(sig)
+	{
+		case SIGCHLD:
+			fprintf(stderr, "): death of packet injector - exiting\n");
+			break;
+		default:
+			common_sig(sig);
+	}
+	exit(sig);
 }
 void pg_sig(int sig)
 {
-	printf("pg: dead %d\n", sig);
-	exit(0);
+	immune2sig();
+	fprintf(stderr, "Packet injector (child");
+	common_sig(sig);
 }
 
+
 int main(int argc, char **argv)
 {
-	char str_src_ip[STR_IP_LEN], str_dst_ip[BUFSIZ];
-	char *ptr, *ptr1, *ptr2;
+	char str_ip[BUFSIZ], *ptr1, *ptr2;
 	const u_char *packet;
 	u_int16_t port, fport=1, lport=1024;
 	u_int32_t ip, net_ip, netmask;
-	int npid, scan_speed=0, cmd_parse, i, j, gotcha[2], known, onlocalnet;
+	long ip_diff;
+	int npid, scan_speed=0, cmd_parse, i, j, gotcha[2], known, onlocalnet, localhost;
 	int i_ip=0, max_retries, max_timeout=0, verbose=0, nbpkt=0, serv_scan=1;
-	struct libnet_link_int *interface;
-	struct pcap *descr;
+	struct libnet_link_int *interface=NULL;
 	struct pcap_pkthdr phdr;
 	struct ether_header *ehdr;
 	struct ether_arp *ahdr, duplic;
@@ -73,15 +100,15 @@ int main(int argc, char **argv)
 	struct tcphdr *thdr;
 	struct timeval tv;
 	sock_elt sckelt;
+	netdev *nd;
 
 
-// ugly debugging code
-for(i=0; i<17; i++) signal(i, main_sig);
+	for(i=0; i<=SIGCHLD; i++) signal(i, main_sig);
 
 
-	// initialize some vars
+	// initialize data
 	rand_seed=libnet_seed_prand();
-	if(!(src_ip=(arp_elt *)malloc(sizeof(arp_elt)))) error(NULL, -2);
+	dev.name[0]='\0';
 	// load known services in memory
 	st=services_load();
 
@@ -96,65 +123,56 @@ for(i=0; i<17; i++) signal(i, main_sig);
 			case 's':
 				// get the src_ip list
 				ptr1=optarg;
-				do
-				{
-					for(ptr=ptr1; *ptr && !isdigit(*ptr); ptr++);
-					if(!*ptr) libnet_error(LIBNET_ERR_FATAL, "Invalid list of source IP addresses: %s\n", optarg);
-					for(ptr1=ptr2=ptr; *ptr1 && *ptr1!=',' && *ptr1!='-'; ptr1++)
-						if(*ptr1=='.') ptr2=ptr1;
-					switch(*ptr1)
-					{
-						case '-':
-						case ',':
-						case 0:
-							if((i=ptr1-ptr)>=STR_IP_LEN)
-								libnet_error(LIBNET_ERR_FATAL, "Invalid list of source IP addresses: %s\n", optarg);
-							strncpy(str_src_ip, ptr, i);
-							str_src_ip[i]='\0';
-							if(!(src_ip=(arp_elt *)realloc(src_ip, (nb_src_ip+1)*sizeof(arp_elt))))
-								error(NULL, -2);
-							if((src_ip[nb_src_ip].ip=libnet_name_resolve(str_src_ip, LIBNET_RESOLVE))==-1)
-								libnet_error(LIBNET_ERR_FATAL, "libnet_name_resolve: invalid src_ip address %s\n", str_src_ip);
-							get_ea(src_ip[nb_src_ip++].ea);
-							if(!*ptr1 || *ptr1==',') break;
-							if(!isdigit(*(++ptr1))) error("Invalid list of src_ip", -1);
-							if((j=atoi(ptr1)-atoi(++ptr2))<0) error("Invalid list of src_ip", -1);
-							if(!(src_ip=(arp_elt *)realloc(src_ip, (nb_src_ip+j)*sizeof(arp_elt))))
-								error(NULL, -2);
-							for(i=0; i<j; i++)
-							{
-								src_ip[nb_src_ip+i].ip=src_ip[nb_src_ip-1].ip+ntohl(i+1);
-								get_ea(src_ip[nb_src_ip+i].ea);
-							}
-							nb_src_ip+=j;
-							for(; *ptr1 && *ptr1!=','; ptr1++);
-							break;
+				j=0; // indicates if we are handling a range of IPs (xxx.xxx.xxx.xxx/yyy.yyy.yyy.yyy)
+				while(*ptr1) {
+					for(ptr2=ptr1; *ptr1 && *ptr1!=',' && *ptr1!='/'; ptr1++);
+					if((i=ptr1-ptr2)>=BUFSIZ) error("Source host name too long", -1);
+					strncpy(str_ip, ptr2, i);
+					str_ip[i]='\0';
+					if((ip=libnet_name_resolve(str_ip, LIBNET_RESOLVE))==-1)
+						libnet_error(LIBNET_ERR_FATAL, "libnet_name_resolve: not able to resolve: %s\n", str_ip);
+					if(j) { // if an IP range is specified
+						j=0;
+						ip_diff=ntohl(ip)-ntohl(src_ip[nb_src_ip-1].ip);
+						if(ip_diff<0) error("Invalid source IP addresses range", -1);
+						if(!(src_ip=(arp_elt *)realloc(src_ip, (ip_diff+nb_src_ip)*sizeof(arp_elt)))) error(NULL, -2);
+						for(i=0; i<ip_diff; i++) {
+							src_ip[nb_src_ip+i].ip=src_ip[nb_src_ip-1].ip+ntohl(i+1);
+							get_ea(src_ip[nb_src_ip+i].ea, rand_seed);
+						}
+						nb_src_ip+=ip_diff;
+					} else {
+						if(!(src_ip=(arp_elt *)realloc(src_ip, (nb_src_ip+1)*sizeof(arp_elt)))) error(NULL, -2);
+						src_ip[nb_src_ip].ip=ip;
+						get_ea(src_ip[nb_src_ip++].ea, rand_seed);
 					}
-				} while(*ptr1);
+					if(!*ptr1) break;
+					if(*ptr1=='/') j=1;
+					ptr1++;
+				}
 				break;
 			case 'd':
-				// dst_ip
-				strncpy(str_dst_ip, optarg, BUFSIZ - 1);
-				str_dst_ip[BUFSIZ-1]='\0';
-				if((dst_ip=libnet_name_resolve(str_dst_ip, LIBNET_RESOLVE))==-1)
-					libnet_error(LIBNET_ERR_FATAL, "libnet_name_resolve: invalid target: %s\n", str_dst_ip);
+				// target host
+				strncpy(str_ip, optarg, BUFSIZ-1);
+				str_ip[BUFSIZ-1]='\0';
+				if((dst_ip=libnet_name_resolve(str_ip, LIBNET_RESOLVE))==-1)
+					libnet_error(LIBNET_ERR_FATAL, "libnet_name_resolve: invalid target: %s\n", str_ip);
 				break;
 			case 'p':
 				// first_port and last_port
 				if(!(fport=atoi(optarg))) error("atoi: invalid first port", -1);
-				for(ptr=optarg; isdigit(*ptr); ptr++);
-				switch(*ptr)
+				for(ptr1=optarg; isdigit(*ptr1); ptr1++);
+				switch(*ptr1)
 				{
 					case 0:
 						lport=fport;
 						break;
 					case '-':
-						if(!(lport=atoi(++ptr))) error("atoi: invalid last port", -1);
+						if(!(lport=atoi(++ptr1))) error("atoi: invalid last port", -1);
 						if(lport<fport) error("Invalid first and last ports", -1);
 						break;
 					default:
 						libnet_error(LIBNET_ERR_FATAL, "Invalid argument: %s\n", optarg);
-						break;
 				}
 				serv_scan=0;
 				break;
@@ -184,8 +202,7 @@ for(i=0; i<17; i++) signal(i, main_sig);
 				else error("Invalid scan_speed", -1);
 				break;
 			case 'e':
-				if(!(ping_port=atoi(optarg)))
-				{
+				if(!(ping_port=atoi(optarg))) {
 					libnet_error(LIBNET_ERR_WARNING, "Invalid ping_port (using port 80)\n");
 					ping_port=80;
 				}
@@ -194,8 +211,7 @@ for(i=0; i<17; i++) signal(i, main_sig);
 				verbose++;
 				break;
 			case 'f':
-				switch(optarg[0])
-				{
+				switch(optarg[0]) {
 					case 't':
 					case 'T':
 						fragment=TINY_FRAG;
@@ -203,34 +219,41 @@ for(i=0; i<17; i++) signal(i, main_sig);
 					case 'o':
 					case 'O':
 						fragment=OVERLAP_FRAG;
-						if(spoofing!=MAC_SPOOF) error("MAC spoofing is required for fragment overlapping", -1);
+						if(spoofing!=MAC_SPOOF)
+						{ spoofing=IP_SPOOF; libnet_error(LIBNET_ERR_WARNING, "Fragment overlapping requires spoofing at physical/link level\n"); }
 						break;
 					default:
 						usage(argv[0]);
-						break;
 				}
 				break;
 			case 'i':
-				eth=optarg;
+				strncpy(dev.name, optarg, IFNAMSIZ);
+				dev.name[IFNAMSIZ-1]='\0';
+				if(!strncmp("lo", dev.name, 2) && spoofing==MAC_SPOOF) {
+					libnet_error(LIBNET_ERR_WARNING, "Scanning on loopback requires spoofing at IP level\n");
+					spoofing=IP_SPOOF;
+				}
 				break;
 			case 'S':
-				if(!strncmp("ip", optarg, 2))
-				{
-					if(fragment==OVERLAP_FRAG) error("MAC spoofing is required for fragment overlapping", -1);
+				if(!strncmp("ip", optarg, 2)) {
+					if(fragment==OVERLAP_FRAG)
+						{ spoofing=MAC_SPOOF; libnet_error(LIBNET_ERR_WARNING, "Fragment overlapping requires spoofing at physical/link level\n"); }
 					else spoofing=IP_SPOOF;
+				} else {
+					spoofing=MAC_SPOOF;
+					if(dev.name[0]) if(!strncmp("lo", dev.name, 2))
+						{ spoofing=IP_SPOOF; libnet_error(LIBNET_ERR_WARNING, "Scanning on loopback requires spoofing at IP level\n"); }
 				}
-				else if(!strncmp("mac", optarg, 3)) spoofing=MAC_SPOOF;
 				break;
 			case '?':
 			default:
 				usage(argv[0]);
 		}
 	}
-	if(!nb_src_ip || !dst_ip) usage(argv[0]);
+	if(!dst_ip) usage(argv[0]);
 
 	// define the number of retries
-	switch(scan_type)
-	{
+	switch(scan_type) {
 		case SYN_SCAN:
 		case ACK_SCAN:
 		case FWRD_SCAN:
@@ -238,15 +261,50 @@ for(i=0; i<17; i++) signal(i, main_sig);
 			break;
 		default:
 			max_retries=10;
-			break;
 	}
 
 
+	// guess if we'll scan localhost
+	if(!(dl=get_localdevs())) libnet_error(LIBNET_ERR_FATAL, "get_localdevs: not able to get the network device list\n");
+	localhost=islocalhost(dl, dst_ip);
+
+	// get a network device name, if none is given
+	if(!dev.name[0]) {
+		if(localhost) {
+			strcpy(dev.name, "lo");
+			if(spoofing!=IP_SPOOF) {
+				libnet_error(LIBNET_ERR_WARNING, "Scanning your localhost requires spoofing at IP level\n");
+				spoofing=IP_SPOOF;
+			}
+			if(fragment==OVERLAP_FRAG) {
+				libnet_error(LIBNET_ERR_WARNING, "Fragment overlapping requires spoofing at physical level, and this is not compliant with scaning through loopback\n");
+				fragment=NO_FRAG;
+			}
+		} else {
+			if((nd=get_suitable_dev(dl, dst_ip))) dev=*nd;
+			else if((ptr1=pcap_lookupdev(errbuf))) {
+				strncpy(dev.name, ptr1, IFNAMSIZ);
+				dev.name[IFNAMSIZ-1]='\0';
+			}
+			else libnet_error(LIBNET_ERR_FATAL, "All attemps to get a suitable network device failed. You should try the -i option to specify the device to use.\n");
+			if(strncmp("et", dev.name, 2)) error("Device %s is not yet supported by GPS.", -1);
+		}
+	}
+	fprintf(stdout, "Network device: %s\n", dev.name);
+
+
+	// if no source IPs have been specified, search for unused IPs
+	if(!src_ip) {
+		if(localhost) error("You must specify a source IP list if you want to scan your localhost", -1);
+		if(verbose) fprintf(stdout, "Looking for IP addresses which are not used on the LAN..\n");
+		src_ip=get_src_ip_list(&dev, ping_port, rand_seed, &nb_src_ip, dst_ip);
+	}
+
+	
 	// allocate memory for data storage and initialize data
 	gotcha[1]=0;
 	nbports=lport-fport+1;
-	if(st && serv_scan)
-	{
+	if(st && serv_scan) {
 		for(i=0; st->tab[i].port<fport && i<st->nb; i++) nbports++;
 		for(i=st->nb-1; st->tab[i].port>lport && i>=0; i--) nbports++;
 	}
@@ -258,40 +316,30 @@ for(i=0; i<17; i++) signal(i, main_sig);
 	if(st && serv_scan) for(j=st->nb-1, i=nbports-1; st->tab[j].port>lport && j>=0; j--, i--) init_port(i, st->tab[j].port);
 
 
-	// get an ethernet device name, if none is specified
-	if(!eth) if(!(eth=pcap_lookupdev(errbuf))) error(errbuf, -1);
-	fprintf(stdout, "Network device: %s\n", eth);
-
-
 	// get the local nework IP, and the netmask
-	if(pcap_lookupnet(eth, &net_ip, &netmask, errbuf))
+	if(pcap_lookupnet(dev.name, &net_ip, &netmask, errbuf))
 		libnet_error(LIBNET_ERR_FATAL, "pcap_lookupnet failed (%s)\n", errbuf);
 	// see if we can reach the host directly
 	onlocalnet=IS_ON_LOCAL_NET(dst_ip, netmask, net_ip);
 
 	// if MAC spoofing is activated, get the MAC address of
 	// the host or gateway to which send the ethernet trames
-	if(spoofing==MAC_SPOOF)
-	{
+	if(spoofing==MAC_SPOOF) {
 		if(verbose) fprintf(stdout, "Searching for the destination MAC address\n");
-		if(onlocalnet)
-		{
-			if(get_mac(FARP, dst_ip, dst_mac, eth)==-1)
-				if(send_ARP_request(eth, &src_ip[0].ip, &dst_ip, src_ip[0].ea, dst_mac)==-1)
+		if(onlocalnet) {
+			if(get_mac(dst_ip, dst_mac, dev.name)==-1)
+				if(send_ARP_request(dev.name, &src_ip[0].ip, &dst_ip, src_ip[0].ea, dst_mac)==-1)
 					libnet_error(LIBNET_ERR_FATAL, "Not able to get the MAC address of host: %s\n", libnet_host_lookup(dst_ip, LIBNET_RESOLVE));
-		}
-		else
-		{
+		} else {
 			// this is not very smartly handled, as far as get_gateway_ip returns
 			// the first gateway IP it finds in FROUTE (/proc/net/route)
 			if((ip=get_gateway_ip(FROUTE))==-1)
 				libnet_error(LIBNET_ERR_FATAL, "get_gateway_ip: not able to find a gateway\n");
-			if(get_mac(FARP, ip, dst_mac, eth)==-1)
-				if(send_ARP_request(eth, &src_ip[0].ip, &ip, src_ip[0].ea, dst_mac)==-1)
+			if(get_mac(ip, dst_mac, dev.name)==-1)
+				if(send_ARP_request(dev.name, &src_ip[0].ip, &ip, src_ip[0].ea, dst_mac)==-1)
 					libnet_error(LIBNET_ERR_FATAL, "Not able to get the MAC address of host: %s\n", libnet_host_lookup(ip, LIBNET_RESOLVE));
 		}
-		if(verbose>2)
-		{
+		if(verbose>1) {
 			fprintf(stdout, "Destination MAC address: ");
 			for(i=0; i<ETH_ALEN; i++) fprintf(stdout, "%x%c", dst_mac[i], (i==ETH_ALEN-1)?'\n':':');
 		}
@@ -300,17 +348,17 @@ for(i=0; i<17; i++) signal(i, main_sig);
 
 	// get a timeout value
 	if(verbose) fprintf(stdout, "Evaluating a timeout value..\n");
-	switch(spoofing)
+	if(localhost) max_timeout=DEFAULT_TIMEOUT_LOOPB;
+	else switch(spoofing)
 	{
 		case IP_SPOOF:
-			max_timeout=ip_eval_timeout(eth, rand_seed, verbose, src_ip[0].ip, dst_ip, ping_port, src_ip[0].ea);
+			max_timeout=ip_eval_timeout(dev.name, rand_seed, verbose, src_ip[0].ip, dst_ip, ping_port, src_ip[0].ea);
 			break;
 		case MAC_SPOOF:
-			max_timeout=ethernet_eval_timeout(eth, rand_seed, verbose, src_ip[0].ip, dst_ip, ping_port, src_ip[0].ea, dst_mac);
+			max_timeout=ethernet_eval_timeout(dev.name, rand_seed, verbose, src_ip[0].ip, dst_ip, ping_port, src_ip[0].ea, dst_mac);
 			break;
 	}
-	if(max_timeout==-1)
-	{
+	if(max_timeout==-1) {
 		if(onlocalnet) max_timeout=DEFAULT_TIMEOUT_LOCAL;
 		else max_timeout=DEFAULT_TIMEOUT_DIST;
 		libnet_error(LIBNET_ERR_WARNING, "Timeout evaluation failure. Are we talking to a dead host ?\n");
@@ -323,12 +371,10 @@ for(i=0; i<17; i++) signal(i, main_sig);
 	// child  thread: packet generator
 	if(socketpair(AF_UNIX, SOCK_STREAM, 0, sockets)<0) error("socketpair: creation failed", -1);
 	if((npid=fork())==-1) error("fork: thread duplication failed", -1);
-	if(!npid) // child
-	{
-// ugly debugging code
-for(i=0; i<20; i++) signal(i, pg_sig);
+	if(!npid) { // child
+		for(i=0; i<=SIGCHLD; i++) signal(i, pg_sig);
 		close(sockets[0]);
-		switch(spoofing)
+		switch(spoofing) 
 		{
 			case IP_SPOOF:
 				ip_packet_generator(scan_speed);
@@ -341,22 +387,20 @@ for(i=0; i<20; i++) signal(i, pg_sig);
 	close(sockets[1]);
 
 	// allocate memory for data
-	if(scan_type!=FWRD_SCAN)
-	{
+	if(scan_type!=FWRD_SCAN) {
 		if(!(tab_response=(u_int16_t *)malloc(nbports*sizeof(u_int16_t)))) error(NULL, -2);
-	}
-	else
-	{
+	} else {
 		if(!(tab_fwrd=(fwrd_elt *)malloc(nb_src_ip*sizeof(fwrd_elt)))) error(NULL, -2);
 		for(i=0; i<nb_src_ip; i++) init_fwrd_elt(&tab_fwrd[i], src_ip[i].ip);
 	}
 
 	// open a raw link interface to send spoofed ARP replies if needed (something like ARP poisoning)
-	if(!(interface=libnet_open_link_interface(eth, errbuf)))
-		libnet_error(LIBNET_ERR_FATAL, "libnet_open_link_interface failed (%s)\n", errbuf);
+	if(!localhost)
+		if(!(interface=libnet_open_link_interface(dev.name, errbuf)))
+			libnet_error(LIBNET_ERR_FATAL, "libnet_open_link_interface failed (%s)\n", errbuf);
 
 	// open network device in promiscuous mode
-	if(!(descr=pcap_open_live(eth, 65536, 1, -1, errbuf)))
+	if(!(descr=pcap_open_live(dev.name, 65536, 1, -1, errbuf)))
 		libnet_error(LIBNET_ERR_FATAL, "pcap_open_live failed (%s)\n", errbuf);
 
 
@@ -377,41 +421,12 @@ for(i=0; i<20; i++) signal(i, pg_sig);
 		if((packet=pcap_next(descr, &phdr)))
 		{
 			// packet filter
-			ehdr=(struct ether_header *)packet;
-			switch(ntohs(ehdr->ether_type))
+			if(localhost) // scaning localhost through lo
 			{
-				case ETHERTYPE_ARP: // ARP packet
-					ahdr=(struct ether_arp *)((u_char *)ehdr + ETHER_HDR_LEN);
-					// break if it is not an ARP request
-					if(ntohs(ahdr->ea_hdr.ar_op)!=ARPOP_REQUEST) break;
-					// break if the ARP request do not concern
-					// the hosts we pretend to be
-					if((j=search_src_ip(*((u_int32_t *)ahdr->arp_tpa)))==-1) break;
-					// we have to reply this ARP request
-					duplic=*ahdr;
-					ahdr->ea_hdr.ar_op=ntohs(ARPOP_REPLY);
-					// build the spoofed ARP reply
-					for(i=0; i<ETH_ALEN; i++) // exchange ARP addresses
-					{
-						ahdr->arp_sha[i]=ehdr->ether_shost[i]=(u_int8_t)src_ip[j].ea[i];
-						ahdr->arp_tha[i]=ehdr->ether_dhost[i]=duplic.arp_sha[i];
-					}
-					for(i=0; i<4; i++) // exchange IP adresses
-					{
-						ahdr->arp_spa[i]=duplic.arp_tpa[i];
-						ahdr->arp_tpa[i]=duplic.arp_spa[i];
-					}
-					// send the fake ARP reply
-					if(libnet_write_link_layer(interface, eth, (u_char *)packet, ETHER_HDR_LEN + sizeof(struct ether_arp))==-1)
-						libnet_error(LIBNET_ERR_WARNING, "libnet_write_link layer failed\n");
-					break;
-				case ETHERTYPE_IP: // IP packet
-					ihdr=(struct ip *)((u_char *)ehdr + ETHER_HDR_LEN);
-					// break if the packet do not come from the target host
-					// or if it is not addressed to the hosts we pretend to be
-					// or if it is fragmented, or if it is not a TCP packet
-					if(ihdr->ip_src.s_addr!=dst_ip || ((scan_type!=FWRD_SCAN) ? (search_src_ip(ihdr->ip_dst.s_addr)==-1) : (ihdr->ip_dst.s_addr!=src_ip[i_ip].ip)) || (ihdr->ip_off & !IP_DF) || ihdr->ip_p!=IPPROTO_TCP) break;
-					thdr=(struct tcphdr *)((u_char *)ihdr + 4 * ihdr->ip_hl);
+				ihdr=(struct ip *)(packet+4);
+				if(ihdr->ip_src.s_addr==dst_ip && ihdr->ip_dst.s_addr==dst_ip && !(ihdr->ip_off & ~ntohs(IP_DF)) && ihdr->ip_p==IPPROTO_TCP)
+				{
+					thdr=(struct tcphdr *)((u_char *)ihdr+4*ihdr->ip_hl);
 					// test packet flags
 					i=0;
 					switch(scan_type)
@@ -419,8 +434,10 @@ for(i=0; i<20; i++) signal(i, pg_sig);
 						case SYN_SCAN:			// SYN scan
 							if(thdr->th_flags!=(TH_SYN|TH_ACK)) i=1;
 							if(thdr->th_flags & TH_RST)
-								if((j=search_port(ntohs(thdr->th_sport)))!=-1)
+								if((j=search_port(ntohs(thdr->th_sport)))!=-1) {
 									tab_port[j].gotcha=1;
+									nbpkt++;
+								}
 							break;
 						case ACK_SCAN:			// ACK scan
 						case FIN_SCAN:			// FIN scan
@@ -442,19 +459,103 @@ for(i=0; i<20; i++) signal(i, pg_sig);
 							nbpkt++;
 							break;
 					}
-					if(i) break;
-					// we have a matching packet, see if we already have it
-					port=ntohs(thdr->th_sport);
-					if((i=search_port(port))!=-1)
+					if(!i)
 					{
-						if(!tab_port[i].gotcha)
+						// we have a matching packet, see if we already have it
+						port=ntohs(thdr->th_sport);
+						if((i=search_port(port))!=-1)
 						{
-							tab_port[i].gotcha=1;
-							tab_response[nbopt++]=port;
+							if(!tab_port[i].gotcha)
+							{
+								tab_port[i].gotcha=1;
+								tab_response[nbopt++]=port;
+							}
 						}
+						nbpkt++;
 					}
-					nbpkt++;
-					break;
+				}
+			}
+			else
+			{
+				ehdr=(struct ether_header *)packet;
+				switch(ntohs(ehdr->ether_type))
+				{
+					case ETHERTYPE_ARP: // ARP packet
+						ahdr=(struct ether_arp *)((u_char *)ehdr + ETHER_HDR_LEN);
+						// break if it is not an ARP request
+						if(ntohs(ahdr->ea_hdr.ar_op)!=ARPOP_REQUEST) break;
+						// break if the ARP request do not concern
+						// the hosts we pretend to be
+						if((j=search_src_ip(*((u_int32_t *)ahdr->arp_tpa)))==-1) break;
+						// we have to reply this ARP request
+						duplic=*ahdr;
+						ahdr->ea_hdr.ar_op=ntohs(ARPOP_REPLY);
+						// build the spoofed ARP reply
+						for(i=0; i<ETH_ALEN; i++) // exchange ARP addresses
+						{
+							ahdr->arp_sha[i]=ehdr->ether_shost[i]=(u_int8_t)src_ip[j].ea[i];
+							ahdr->arp_tha[i]=ehdr->ether_dhost[i]=duplic.arp_sha[i];
+						}
+						for(i=0; i<4; i++) // exchange IP adresses
+						{
+							ahdr->arp_spa[i]=duplic.arp_tpa[i];
+							ahdr->arp_tpa[i]=duplic.arp_spa[i];
+						}
+						// send the fake ARP reply
+						if(libnet_write_link_layer(interface, dev.name, (u_char *)packet, ETHER_HDR_LEN + sizeof(struct ether_arp))==-1)
+							libnet_error(LIBNET_ERR_WARNING, "libnet_write_link layer failed\n");
+						break;
+					case ETHERTYPE_IP: // IP packet
+						ihdr=(struct ip *)((u_char *)ehdr + ETHER_HDR_LEN);
+						// break if the packet do not come from the target host
+						// or if it is not addressed to the hosts we pretend to be
+						// or if it is fragmented, or if it is not a TCP packet
+						if(ihdr->ip_src.s_addr!=dst_ip || ((scan_type!=FWRD_SCAN) ? (search_src_ip(ihdr->ip_dst.s_addr)==-1) : (ihdr->ip_dst.s_addr!=src_ip[i_ip].ip)) || (ihdr->ip_off & ~ntohs(IP_DF)) || ihdr->ip_p!=IPPROTO_TCP) break;
+						thdr=(struct tcphdr *)((u_char *)ihdr + 4 * ihdr->ip_hl);
+						// test packet flags
+						i=0;
+						switch(scan_type)
+						{
+							case SYN_SCAN:			// SYN scan
+								if(thdr->th_flags!=(TH_SYN|TH_ACK)) i=1;
+								if(thdr->th_flags & TH_RST)
+									if((j=search_port(ntohs(thdr->th_sport)))!=-1)
+										tab_port[j].gotcha=1;
+								break;
+							case ACK_SCAN:			// ACK scan
+							case FIN_SCAN:			// FIN scan
+							case XMAS_SCAN:			// Xmas Tree scan
+							case NULL_SCAN:			// NULL scan
+							case RAND_SCAN:			// RAND scan
+								if(!(thdr->th_flags & TH_RST)) i=1;
+								break;
+							case FWRD_SCAN:			// FWRD scan
+								if(thdr->th_flags & TH_RST) i=1;
+								// we have a matching packet, see if we already have it
+								port=ntohs(thdr->th_sport);
+								if(search_fwrd_port(&tab_fwrd[i_ip], port)==-1)
+								{
+									nbopt++;
+									add_fwrd_port(&tab_fwrd[i_ip], port);
+									if((j=search_port(port))!=-1) tab_port[j].gotcha=1;
+								}
+								nbpkt++;
+								break;
+						}
+						if(i) break;
+						// we have a matching packet, see if we already have it
+						port=ntohs(thdr->th_sport);
+						if((i=search_port(port))!=-1)
+						{
+							if(!tab_port[i].gotcha)
+							{
+								tab_port[i].gotcha=1;
+								tab_response[nbopt++]=port;
+							}
+						}
+						nbpkt++;
+						break;
+				}
 			}
 		}
 
@@ -470,6 +571,7 @@ for(i=0; i<20; i++) signal(i, pg_sig);
 				}
 				break;
 			case -2: // all packets have been sent
+				signal(SIGCHLD, SIG_IGN);
 				continue;
 			case -3: // ready to send packets from next IP (FWRD mode)
 				break;
@@ -489,7 +591,11 @@ for(i=0; i<20; i++) signal(i, pg_sig);
 			{
 				if(!tab_port[i].gotcha)
 				{
-					if(tab_port[i].nbretries>max_retries) tab_port[i].gotcha=1;
+					if(tab_port[i].nbretries>max_retries) {
+						tab_port[i].gotcha=1;
+						if(localhost && scan_type==SYN_SCAN)
+							tab_response[nbopt++]=tab_port[i].port;
+					}
 					else
 					{
 						gotcha[0]=0;
@@ -508,7 +614,7 @@ for(i=0; i<20; i++) signal(i, pg_sig);
 
 			if(gotcha[0]) // all packets have been received; inform the child thread
 			{
-				if(verbose>0)
+				if(verbose)
 				{
 					fprintf(stdout, "All packets");
 					if(scan_type==FWRD_SCAN)
@@ -519,6 +625,7 @@ for(i=0; i<20; i++) signal(i, pg_sig);
 				sckelt.i=-1;
 				i=0; while(write(sockets[0], &sckelt, sizeof(sock_elt))<0 && i++<3);
 				while(sckelt.i>-2) read(sockets[0], &sckelt, sizeof(sock_elt));
+				if(sckelt.i==-2) signal(SIGCHLD, SIG_IGN);
 			}
 		}
 
@@ -585,9 +692,9 @@ for(i=0; i<20; i++) signal(i, pg_sig);
 						for(i=0; i<nbopt; i++)
 						{
 							if((known=services_known(tab_response[i], st))==-1)
-								fprintf(stdout, "Port %u is open\n", tab_response[i]);
+								fprintf(stdout, "Port %5u is open\n", tab_response[i]);
 							else // port service detail
-								fprintf(stdout, "Port %u: %s (%s) is open\n", tab_response[i], st->tab[known].service, (st->tab[known].comment[0]) ? st->tab[known].comment : "-");
+								fprintf(stdout, "Port %5u: %s (%s) is open\n", tab_response[i], st->tab[known].service, (st->tab[known].comment[0]) ? st->tab[known].comment : "-");
 						}
 					}
 					break;
@@ -611,7 +718,7 @@ for(i=0; i<20; i++) signal(i, pg_sig);
 						fprintf(stdout, "Ports: ");
 						if(nbopt<nbports/2)
 						{
-							for(i=0; i<nbopt; i++) fprintf(stdout, "%u, ", tab_response[i]);
+							for(i=0; i<nbopt; i++) fprintf(stdout, "%5u, ", tab_response[i]);
 							fprintf(stdout, "\b\b are _not_ filtered\n");
 						}
 						else
@@ -623,7 +730,7 @@ for(i=0; i<20; i++) signal(i, pg_sig);
 								else j++;
 							}
 							for(; i<nbports; i++)
-								fprintf(stdout, "%d, ", tab_port[i].port);
+								fprintf(stdout, "%5d, ", tab_port[i].port);
 							fprintf(stdout, "\b\b are filtered\n");
 						}
 					}
@@ -641,18 +748,20 @@ for(i=0; i<20; i++) signal(i, pg_sig);
 		free(tab_fwrd);
 	}
 	else free(tab_response);
+	if(dl) free_devlist(dl);
 	free(tab_port);
 	free(src_ip);
 	pcap_close(descr);
-	if(libnet_close_link_interface(interface)==-1)
-		libnet_error(LIBNET_ERR_WARNING, "libnet_close_link_interface failed\n");
-	exit(0);
+	if(!localhost)
+		if(libnet_close_link_interface(interface)==-1)
+			libnet_error(LIBNET_ERR_WARNING, "libnet_close_link_interface failed\n");
+	return(0);
 }
 
-void get_ea(char *ea)
+void get_ea(char *ea, int rs)
 {
 	int i;
-	for(i=0; i<ETH_ALEN; i++) ea[i]=(u_char)libnet_get_prand(rand_seed);
+	for(i=0; i<ETH_ALEN; i++) ea[i]=(u_char)libnet_get_prand(rs);
 }
 
 void init_port(int i, u_int16_t port)
@@ -691,16 +800,16 @@ void reverse_scan_disp()
 		if(tab_port[i].port<tab_response[j])
 		{
 			if((known=services_known(tab_port[i].port, st))==-1)
-				fprintf(stdout, "Port %d is open or filtered\n", tab_port[i].port);
-			else fprintf(stdout, "Port %d: %s (%s) is open (or filtered)\n", tab_port[i].port, st->tab[known].service, (st->tab[known].comment[0]) ? st->tab[known].comment : "-");
+				fprintf(stdout, "Port %5d  is open or filtered\n", tab_port[i].port);
+			else fprintf(stdout, "Port %5d: %s (%s) is open (or filtered)\n", tab_port[i].port, st->tab[known].service, (st->tab[known].comment[0]) ? st->tab[known].comment : "-");
 		}
 		else j++;
 	}
 	for(; i<nbports; i++)
 	{
 		if((known=services_known(tab_port[i].port, st))==-1)
-			fprintf(stdout, "%d\n", tab_port[i].port);
-		else fprintf(stdout, "Port %d: %s (%s) is open (or fileterd)\n", tab_port[i].port, st->tab[known].service, (st->tab[known].comment[0]) ? st->tab[known].comment : "-");
+			fprintf(stdout, "%5d\n", tab_port[i].port);
+		else fprintf(stdout, "Port %5d: %s (%s) is open (or fileterd)\n", tab_port[i].port, st->tab[known].service, (st->tab[known].comment[0]) ? st->tab[known].comment : "-");
 	}
 }
 
@@ -873,7 +982,9 @@ void ip_packet_generator(int scan_speed)
 	free(tab_switch_r2o);
 	free(tab_port);
 	free(src_ip);
+	if(dl) free_devlist(dl);
 
+	sleep(1);
 	exit(0);
 }
 
@@ -897,7 +1008,7 @@ void ethernet_packet_generator(int scan_speed)
 	i=0; while(read(sockets[1], &sckelt, sizeof(sock_elt))<0 && i++<3);
 
 	// open a link layer interface
-	if(!(interface=libnet_open_link_interface(eth, errbuf)))
+	if(!(interface=libnet_open_link_interface(dev.name, errbuf)))
 		libnet_error(LIBNET_ERR_FATAL, "libnet_open_link_interface failed (%s)\n", errbuf);
 
 	// set up TCP flags for fwrd scan
@@ -935,7 +1046,7 @@ void ethernet_packet_generator(int scan_speed)
 					libnet_build_tcp(tab_port[i].src_port, tab_port[i].port, tab_port[i].seq, tab_port[i].ack, th_flags, 1024, 0, NULL, 0, packets+offset+GPS_ETH_H+GPS_IP_H);
 					// if the checksums are ok, send it
 					if(libnet_do_checksum(packets+offset+GPS_ETH_H, IPPROTO_TCP, GPS_TCP_H)!=-1 && libnet_do_checksum(packets+offset+GPS_ETH_H, IPPROTO_IP, GPS_IP_H)!=-1)
-						libnet_write_link_layer(interface, eth, packets+offset, GPS_TRAME);
+						libnet_write_link_layer(interface, dev.name, packets+offset, GPS_TRAME);
 					break;
 				case TINY_FRAG: // build two tiny fragments
 					// build the first fragment
@@ -951,8 +1062,8 @@ void ethernet_packet_generator(int scan_speed)
 					if(libnet_do_checksum(frag_packets+offset+GPS_ETH_H, IPPROTO_IP, GPS_IP_H)==-1)
 						libnet_error(LIBNET_ERR_FATAL, "libnet_do_checksum failed\n");
 					// send them
-					libnet_write_link_layer(interface, eth, frag_packets+offset, GPS_TRAME-16);
-					libnet_write_link_layer(interface, eth, packets+offset, GPS_TRAME-4);
+					libnet_write_link_layer(interface, dev.name, frag_packets+offset, GPS_TRAME-16);
+					libnet_write_link_layer(interface, dev.name, packets+offset, GPS_TRAME-4);
 					break;
 				case OVERLAP_FRAG: // NOTE: most of recent OSes drop IP
 						   // datagrams with PROTO=TCP and OFFSET=1
@@ -975,8 +1086,8 @@ void ethernet_packet_generator(int scan_speed)
 					thdr=(struct tcphdr *)(packets+offset+GPS_ETH_H+GPS_IP_H);
 					thdr->th_flags=TH_NULL;
 					// send them
-					libnet_write_link_layer(interface, eth, packets+offset, GPS_TRAME-4);
-					libnet_write_link_layer(interface, eth, frag_packets+offset, GPS_TRAME-8);
+					libnet_write_link_layer(interface, dev.name, packets+offset, GPS_TRAME-4);
+					libnet_write_link_layer(interface, dev.name, frag_packets+offset, GPS_TRAME-8);
 					break;
 			}
 
@@ -1006,15 +1117,15 @@ void ethernet_packet_generator(int scan_speed)
 			switch(fragment)
 			{
 				case NO_FRAG:
-					libnet_write_link_layer(interface, eth, packets+offset, GPS_TRAME);
+					libnet_write_link_layer(interface, dev.name, packets+offset, GPS_TRAME);
 					break;
 				case TINY_FRAG:
-					libnet_write_link_layer(interface, eth, frag_packets+offset, GPS_TRAME-16);
-					libnet_write_link_layer(interface, eth, packets+offset, GPS_TRAME-4);
+					libnet_write_link_layer(interface, dev.name, frag_packets+offset, GPS_TRAME-16);
+					libnet_write_link_layer(interface, dev.name, packets+offset, GPS_TRAME-4);
 					break;
 				case OVERLAP_FRAG:
-					libnet_write_link_layer(interface, eth, packets+offset, GPS_TRAME-4);
-					libnet_write_link_layer(interface, eth, frag_packets+offset, GPS_TRAME-8);
+					libnet_write_link_layer(interface, dev.name, packets+offset, GPS_TRAME-4);
+					libnet_write_link_layer(interface, dev.name, frag_packets+offset, GPS_TRAME-8);
 					break;
 			}
 			// inform the sniffer that the packet(s) is/are sent
@@ -1049,17 +1160,20 @@ void ethernet_packet_generator(int scan_speed)
 	free(tab_switch_r2o);
 	free(tab_port);
 	free(src_ip);
+	if(dl) free_devlist(dl);
 
+	sleep(1);
 	exit(0);
 }
 
 
 void usage(char *cmd)
 {
-fprintf(stderr, "Usage: %s -s src_ip[,src_ip_1[,src_ip_2-4..]] -d target [-t scan_type] [-v]\n\
-	   [-r scan_speed] [-p first_port-last_port] [-k 0 | 1] [-e ping_port]\n\
-	   [-f t | o] [-i interface] [-S mac | ip]\n\
- -s src_ip[,src_ip_1..]: list of IP addresses of the hosts we pretend to be\n\
+fprintf(stderr, "Usage: %s -s host1[,host2/host3..] -d target [-t scan_type] [-v]\n\
+	   [-r scan_speed] [-p first_port-last_port] [-k 0 | 1]\n\
+	   [-e ping_port]  [-f t | o] [-i interface] [-S mac | ip]\n\
+ -s host1[,host2/host3]: list of hosts we pretend to be\n\
+			 (use '/' to specify IP ranges)\n\
  -d target.............: target host's IP/name\n\
  -t scan_type..........: stealth scan mode (default: syn)\n\
  			 (syn | xmas | null | fin | ack | rand | fwrd)\n\
@@ -1071,7 +1185,7 @@ fprintf(stderr, "Usage: %s -s src_ip[,src_ip_1[,src_ip_2-4..]] -d target [-t sca
  -v....................: verbose (use twice for more verbose)\n\
  -f t | o..............: fragment IP datagrams (default: no frag)\n\
  			 (t: tiny frags | o: frag overlapping)\n\
- -i interface..........: ethernet interface to use\n\
+ -i interface..........: network interface to use\n\
  -S mac | ip...........: spoofing level (IP or ethernet/MAC; default: mac)\n", cmd);
 exit(-1);
 }
@@ -1081,7 +1195,7 @@ void error(char *msg, int ret)
 	switch(ret)
 	{
 		case -2:
-			fprintf(stderr, "malloc: memory allocation failed\n");
+			fprintf(stderr, "malloc/realloc: memory allocation failed\n");
 			exit(-2);
 		default:
 			fprintf(stderr, "%s\n", msg);
